
/*
 * Copyright 1994 John L. Cwikla
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appears in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of John L. Cwikla or
 * Wolfram Research, Inc not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written
 * prior permission.    John L. Cwikla and Wolfram Research, Inc make no
 * representations about the suitability of this software for any
 * purpose. It is provided "as is" without express or implied warranty.
 *
 * John L. Cwikla and Wolfram Research, Inc disclaim all warranties with
 * regard to this software, including all implied warranties of
 * merchantability and fitness, in no event shall John L. Cwikla or
 * Wolfram Research, Inc be liable for any special, indirect or
 * consequential damages or any damages whatsoever resulting from loss of
 * use, data or profits, whether in an action of contract, negligence or
 * other tortious action, arising out of or in connection with the use or
 * performance of this software.
 *
 * Author:
 *  John L. Cwikla
 *  X Programmer
 *  Wolfram Research Inc.
 *
 *  cwikla@wri.com
*/

/*
** SmartMessageB.c
*/

#include <X11/IntrinsicP.h>
#include <X11/CoreP.h>

#include <Xm/Xm.h>
#include <Xm/BulletinBP.h>
#include "SmartMBP.h"

#define managerTranslations _XmBulletinB_defaultTranslations
extern void _XmBulletinBoardSetDefaultShadow(); /* Why rewrite? */

#define CORE(a) ((a)->core)
#define COMP(a) ((a)->composite)
#define MANAGER(a) ((a)->manager)
#define SMB(a) ((a)->smart_message_box)
#define LABEL(a) ((a)->smart_message_box.labelW)
#define CONTROL(a) ((a)->smart_message_box.controlW)
#define SEP(a) ((a)->smart_message_box.separatorW)

#define MCLASS(a) ((a)->smart_message_box_class)

#define GOOD(a) ((a) && XtIsManaged(a))

#define CH_TYPE(a) (((XmSmartMessageBoxConstraint)((a)->core.constraints))->smart_message_box.childType)

#define IS_ACTION(a) (CH_TYPE(a) == XmCHILD_ACTION)
#define IS_CONTROL(a) (CH_TYPE(a) == XmCHILD_CONTROL)
#define IS_SEPARATOR(a) (CH_TYPE(a) == XmCHILD_SEPARATOR)
#define IS_LABEL(a) (CH_TYPE(a) == XmCHILD_LABEL)

#ifdef _NO_PROTO
static XtProc classInit();
static XtInitProc initialize();
static Boolean setValues();
static Boolean constraintSetValues();
static XtWidgetProc insertChild();
static XtWidgetProc deleteChild();
static Boolean noGeoRequest();
static XmGeoMatrix geoMatrixCreate();
static Boolean cvtStringToSMBChildType();
#else
static XtProc classInit();
static XtInitProc initialize(XmSmartMessageBoxWidget _request, XmSmartMessageBoxWidget _new, String *_args, Cardinal *_numArgs);
static Boolean setValues(XmSmartMessageBoxWidget _current, XmSmartMessageBoxWidget _request, XmSmartMessageBoxWidget _new);
static Boolean constraintSetValues(Widget _current, Widget _request, Widget _new);
static XtWidgetProc insertChild(Widget _w);
static XtWidgetProc deleteChild(Widget _w);
static Boolean noGeoRequest(XmGeoMatrix _geoSpec);
static XmGeoMatrix geoMatrixCreate(Widget _w, Widget _from, XtWidgetGeometry *_pref);
static Boolean cvtStringToSMBChildType(Display *_display, XrmValuePtr _args,
	Cardinal *_numArgs, XrmValuePtr _from, XrmValuePtr _to, XtPointer *_data);
#endif /* _NO_PROTO */

#define LOCALCONSTRAINTREC XmSmartMessageBoxConstraintRec

#define TheOffset(field) XtOffset(XmSmartMessageBoxWidget, smart_message_box.field)
static XtResource resources[] =
{
	{XmNdata, XmCData, XmRPointer, sizeof(XtPointer),
		TheOffset(data), XmRImmediate, (XtPointer)NULL},
	{XmNminimizeButtons, XmCMinimizeButtons, XmRBoolean, sizeof(Boolean),
		TheOffset(minimizeButtons), XmRImmediate, (XtPointer)False},
};

#undef TheOffset

#define TheOffset(field) XtOffset(XmSmartMessageBoxConstraint, smart_message_box.field)
static XtResource constraintResources[] =
{
	{XmNchildType, XmCChildType, XmRSMBChildType, sizeof(unsigned char),
		TheOffset(childType), XmRImmediate, (XtPointer)XmCHILD_ACTION},
};
#undef TheOffset


XmSmartMessageBoxClassRec xmSmartMessageBoxClassRec =
{
	{
		(WidgetClass)&xmBulletinBoardClassRec,				/* superclass */
		"XmSmartMessageBox",								 /* class_name */
		(Cardinal)sizeof(XmSmartMessageBoxRec),						/* widget size */
		(XtProc)classInit,										/* class_init */
		(XtWidgetClassProc)NULL,										/* class_part_init */
		(XtEnum)FALSE,										 /* class_inited */
		(XtInitProc)initialize,									/* initialize */
		(XtArgsProc)NULL,										/* init_hook */
		XtInheritRealize,							/* realize */
		(XtActionList)NULL,										/* actions */
		(Cardinal)0,											 /* num_actions */
		(XtResourceList)resources,									 /* resources */
		(Cardinal)XtNumber(resources),						 /* num_resources */
		NULLQUARK,									 /* xrm_class */
		FALSE,										 /* compress_motion */
		(XtEnum)FALSE,										 /* compress_exposur */
		FALSE,										 /* compress enterleave */
		FALSE,										/* visibility_interest */
		(XtWidgetProc)NULL,										/* destroy */
		(XtWidgetProc)XtInheritResize,
		XtInheritExpose,
		(XtSetValuesFunc)setValues,									 /* set_values */
		(XtArgsFunc)NULL,										/* set_values_hook */
		XtInheritSetValuesAlmost,					/* set_values_almost */
		(XtArgsProc)NULL,										/* get_values_hook */
		XtInheritAcceptFocus,						/* accept_focus */
		XtVersion,									 /* version */
		(XtPointer)NULL,										/* callback_private */
		XtInheritTranslations,
		(XtGeometryHandler)XtInheritQueryGeometry,			/* query_geometry */
		XtInheritDisplayAccelerator,				 /* display_accelerator */
		(XtPointer)NULL,										/* extension */
	},
	{ /* Composite Part */
		(XtGeometryHandler)XtInheritGeometryManager,							 /* geometry_manager */
		(XtWidgetProc)XtInheritChangeManaged,								 /* change_managed */
		(XtWidgetProc)insertChild,						/* inherit_child */
		(XtWidgetProc)deleteChild,						/* delete_child */
		NULL,										/* extension */
	},
	{	/* Constraint Part */
		(XtResourceList)constraintResources,
		XtNumber(constraintResources),
		(Cardinal)sizeof(XmSmartMessageBoxConstraintRec),
		(XtInitProc)NULL,
		(XtWidgetProc)NULL,
		(XtSetValuesFunc)constraintSetValues,
		(XtPointer)NULL,
	},
	{	/* XmManager Part */
		XmInheritTranslations,
		NULL,
		0,	
		NULL,
		0,	
		XmInheritParentProcess,
		NULL,
	},
	{ /* XmBulletinBoard Part */
		TRUE,
		geoMatrixCreate,
		XmInheritFocusMovedProc,	
		NULL,
	},
	{ /* XmSmartMessageBox Part */
		0,
	},
};

WidgetClass xmSmartMessageBoxWidgetClass = (WidgetClass)&xmSmartMessageBoxClassRec;

static XtProc classInit()
{
	XtSetTypeConverter(XtRString, XmRSMBChildType, cvtStringToSMBChildType,
		(XtConvertArgList)NULL, 0,
			XtCacheAll, (XtDestructor)NULL);
}

#ifdef _NO_PROTO
static XtInitProc initialize(_request, _new, _args, _numArgs)
XmSmartMessageBoxWidget _request;
XmSmartMessageBoxWidget _new;
String *_args;
Cardinal *_numArgs;
#else
static XtInitProc initialize(XmSmartMessageBoxWidget _request, XmSmartMessageBoxWidget _new, String *_args, Cardinal *_numArgs)
#endif /* _NO_PROTO */
{
	LABEL(_new) = (Widget)NULL;
	SEP(_new) = (Widget)NULL;
	CONTROL(_new) = (Widget)NULL;
}

#ifdef _NO_PROTO
static XtWidgetProc insertChild(_w)
Widget _w;
#else
static XtWidgetProc insertChild(Widget _w)
#endif /* _NO_PROTO */
{
	XmSmartMessageBoxWidget mb;
	Boolean warning;
	char buffer[255];
	Cardinal numParms;

	mb = (XmSmartMessageBoxWidget)XtParent(_w);

	(*((CompositeWidgetClass)(xmSmartMessageBoxWidgetClass->core_class.superclass))->composite_class.insert_child)(_w);

	if (IS_CONTROL(_w))
	{
		if (CONTROL(mb) == (Widget)NULL)
			CONTROL(mb) = _w;
		else
		{
			numParms = 1;
			sprintf(buffer, "Trying to add %s as another control child widget!\n",
				XtName(_w));

			XtAppWarningMsg(XtWidgetToApplicationContext((Widget)_w),
				"childType", "CHILD_CONTROL", "TooManyControlWidgets",
				buffer, (String *)NULL, &numParms);

			CH_TYPE(_w) = XmCHILD_ACTION;
		}
	}
	else
	if (IS_SEPARATOR(_w))
	{
		if (SEP(mb) == (Widget)NULL)
			SEP(mb) = _w;
		else
		{
			numParms = 1;
			sprintf(buffer, "Trying to add %s as another separator child widget!\n",
				XtName(_w));

			XtAppWarningMsg(XtWidgetToApplicationContext((Widget)_w),
				"childType", "CHILD_SEPARATOR", "TooManySeparatorWidgets",
				buffer, (String *)NULL, &numParms);

			CH_TYPE(_w) = XmCHILD_ACTION;
		}
	}
	else
	if (IS_LABEL(_w))
	{
		if (LABEL(mb) == (Widget)NULL)
			LABEL(mb) = _w;
		else
		{
			numParms = 1;
			sprintf(buffer, "Trying to add %s as another label child widget!\n",
				XtName(_w));

			XtAppWarningMsg(XtWidgetToApplicationContext((Widget)_w),
				"childType", "CHILD_LABEL", "TooManyLabelWidgets",
				buffer, (String *)NULL, &numParms);

			CH_TYPE(_w) = XmCHILD_ACTION;
		}
	}
}

#ifdef _NO_PROTO
static XtWidgetProc deleteChild(_w)
Widget _w;
#else
static XtWidgetProc deleteChild(Widget _w)
#endif /* _NO_PROTO */
{
	XmSmartMessageBoxWidget mb;
	String params[1];
	Cardinal numParms;

	mb = (XmSmartMessageBoxWidget)XtParent(_w);

	if (CONTROL(mb) == _w)
		CONTROL(mb) = (Widget)NULL;
	else
	if (LABEL(mb) == _w)
		LABEL(mb) = (Widget)NULL;
	else
	if (SEP(mb) == _w)
		SEP(mb) = (Widget)NULL;

	(*((CompositeWidgetClass)(xmSmartMessageBoxWidgetClass->core_class.superclass))->composite_class.delete_child)(_w);
}

#ifdef _NO_PROTO
static Boolean setValues(_current, _request, _new)
XmSmartMessageBoxWidget _current;
XmSmartMessageBoxWidget _request;
XmSmartMessageBoxWidget _new;
#else
static Boolean setValues(XmSmartMessageBoxWidget _current, XmSmartMessageBoxWidget _request, XmSmartMessageBoxWidget _new)
#endif /* NO_PROTO */
{
	Boolean redo = FALSE;

	BB_InSetValues(_current) = TRUE;

	if (SMB(_new).minimizeButtons != SMB(_current).minimizeButtons)
		redo = TRUE;
	else
	if (CONTROL(_new) != CONTROL(_current))
		redo = TRUE;
	else
	if (LABEL(_new) != LABEL(_current))
		redo = TRUE;
	else
	if (SEP(_new) != SEP(_current))
		redo = TRUE;

	BB_InSetValues(_current) = FALSE;

	if(redo && (XtClass(_new) == xmSmartMessageBoxWidgetClass))
	{
		_XmBulletinBoardSizeUpdate( (Widget) _new);
		return FALSE;
	}

	return redo;
}

#ifdef _NO_PROTO
static Boolean constraintSetValues(_current, _request, _new)
Widget _current;
Widget _request;
Widget _new;
#else
static Boolean constraintSetValues(Widget _current, Widget _request, Widget _new)
#endif /* _NO_PROTO */
{
	XmSmartMessageBoxWidget smbw;
	Boolean redo = FALSE;

	smbw = (XmSmartMessageBoxWidget)XtParent(_new);

	BB_InSetValues(smbw) = TRUE;

	if (CH_TYPE(_new) != CH_TYPE(_current))
	{
		redo = TRUE;
		if (IS_CONTROL(_new))
			CONTROL(smbw) = _new;
		else
		if (IS_LABEL(_new))
			LABEL(smbw) = _new;
		else
		if (IS_SEPARATOR(_new))
			SEP(smbw) = _new;
	}

	BB_InSetValues(smbw) = FALSE;

	if(redo && (XtClass(smbw) == xmSmartMessageBoxWidgetClass))
	{
		_XmBulletinBoardSizeUpdate( (Widget)smbw);
		return FALSE;
	}

	return redo;
}

#ifdef _NO_PROTO
static XmGeoMatrix geoMatrixCreate(_w, _from, _pref)
Widget _w;
Widget _from;
XtWidgetGeometry *_pref;
#else
static XmGeoMatrix geoMatrixCreate(Widget _w, Widget _from, XtWidgetGeometry *_pref)
#endif /* _NO_PROTO */
{
	XmSmartMessageBoxWidget smbw = (XmSmartMessageBoxWidget)_w;
	XmGeoMatrix geoSpec;
	register XmGeoRowLayout	layoutPtr;
	register XmKidGeometry boxPtr;
	XmKidGeometry tempBox;
	Cardinal actionKids;
	Widget controlKid, curChild;
	Cardinal i, numKids;
	Boolean newRow;
	int nrows;

	nrows = 0;
	actionKids = 0;
	controlKid = (Widget)NULL;

	numKids = COMP(smbw).num_children;

	for(i=0;i<numKids;i++)
	{
		curChild = COMP(smbw).children[i];

		if (IS_ACTION(curChild) && GOOD(curChild))
			actionKids++;
		else
		if (IS_CONTROL(curChild) && GOOD(curChild))
			controlKid = curChild;
	}

	if (actionKids != 0)
		nrows++;

	if (GOOD(LABEL(smbw)) || (controlKid != (Widget)NULL))
		nrows++;

	if (GOOD(SEP(smbw)))
		nrows++;

	geoSpec = _XmGeoMatrixAlloc(nrows, numKids, 0);
	geoSpec->composite = (Widget)smbw;
	geoSpec->instigator = (Widget)_from;
	if (_pref)
		geoSpec->instig_request = *_pref;

	geoSpec->margin_w = BB_MarginWidth(smbw) + MANAGER(smbw).shadow_thickness;
	geoSpec->margin_h = BB_MarginHeight(smbw) + MANAGER(smbw).shadow_thickness;
	geoSpec->no_geo_request = noGeoRequest;

	layoutPtr = &(geoSpec->layouts->row);
	boxPtr = geoSpec->boxes;

	tempBox = boxPtr;
	newRow = FALSE;

	if ( GOOD(LABEL(smbw)) && _XmGeoSetupKid(boxPtr, LABEL(smbw)))
	{
		newRow = TRUE;
		boxPtr++;
	}

	if (controlKid && _XmGeoSetupKid(boxPtr, controlKid))
	{
		newRow = TRUE;
		boxPtr++;
	}

	if ( (boxPtr != tempBox) && controlKid)
	{
		layoutPtr->fill_mode = XmGEO_EXPAND;
		layoutPtr->even_width = 0;
		layoutPtr->stretch_height = 1;
		layoutPtr->space_above = 0; /* BB_MarginHeight(smbw); */
		layoutPtr->space_between = BB_MarginWidth(smbw);
		newRow = TRUE;
	}

	if (newRow)
	{
		boxPtr++;
		layoutPtr++;
	}


	if(	GOOD(SEP(smbw)) && _XmGeoSetupKid( boxPtr, SEP(smbw)))
	{	 
		layoutPtr->fix_up = _XmSeparatorFix;
		layoutPtr->space_above = BB_MarginHeight(smbw)/2;
		boxPtr += 2;
		layoutPtr++;
	}

	if (actionKids)
	{
		for(i=0;i<numKids;i++)
		{
			curChild = COMP(smbw).children[i];

			if (IS_ACTION(curChild) && GOOD(curChild))
				_XmGeoSetupKid( boxPtr++, curChild);
		}

		layoutPtr->fill_mode = XmGEO_CENTER;
		layoutPtr->fit_mode = XmGEO_AVERAGING;
		layoutPtr->even_width = SMB(smbw).minimizeButtons ? 0 : 1;
		layoutPtr->even_height = 1;
		layoutPtr->space_above = BB_MarginHeight(smbw);
		layoutPtr++;
	}

	layoutPtr->space_above = BB_MarginHeight(smbw);
	layoutPtr->end = TRUE;
	return(geoSpec);
}

#ifdef _NO_PROTO
static Boolean noGeoRequest(_geoSpec)
XmGeoMatrix _geoSpec;
#else
static Boolean noGeoRequest(XmGeoMatrix _geoSpec)
#endif /* _NO_PROTO */
{
	if(BB_InSetValues(_geoSpec->composite) && 
		(XtClass(_geoSpec->composite) == xmSmartMessageBoxWidgetClass))
			return(TRUE);

	return( FALSE);
}

#ifdef _NO_PROTO
static void toLower(_str1, _str2, _length)
char *_str1;
char *_str2;
int _length;
#else
static void toLower(char *_str1, char *_str2, int _length)
#endif /* _NO_PROTO */
{
	int i;
	char *ptr;

	for(ptr=_str1,i=0;(ptr!=NULL) && (i<_length);ptr++,i++)
		*(_str2+i) = tolower(*ptr);
}

#ifdef _NO_PROTO
static Boolean cvtStringToSMBChildType(_display, _args, _numArgs, _from, _to, _data)
Display *_display;
XrmValuePtr _args;
Cardinal *_numArgs;
XrmValuePtr _from;
XrmValuePtr _to;
XtPointer *_data;
#else
static Boolean cvtStringToSMBChildType(Display *_display, XrmValuePtr _args,
    Cardinal *_numArgs, XrmValuePtr _from, XrmValuePtr _to, XtPointer *_data)
#endif /* _NO_PROTO */
{
    char *lower;
    static unsigned char childType;
    Boolean badConversion = FALSE;

    if (*_numArgs != 0)
    {
        XtAppWarningMsg(XtDisplayToApplicationContext(_display), "cvtStringToSMBChildType", "wrongParamaters",
        "ResourceError",
        "cvtStringToSMBChildType needs no arguments.",
        (String *)NULL, (Cardinal *)NULL);
    }

    lower = XtNewString(_from->addr);
    toLower(_from->addr, lower, strlen(_from->addr));

    childType = XmCHILD_ACTION;

    if (!strncmp(lower, "control", 7))
        childType = XmCHILD_CONTROL;
    else
    if (!strncmp(lower, "separator", 9))
        childType = XmCHILD_SEPARATOR;
    else
    if (!strncmp(lower, "label", 5))
        childType = XmCHILD_LABEL;
    else
    if (!strncmp(lower, "action", 6))
        childType = XmCHILD_ACTION;
    else
    if (!strncmp(lower, "child_control", 13))
        childType = XmCHILD_CONTROL;
    else
    if (!strncmp(lower, "child_separator", 15))
        childType = XmCHILD_SEPARATOR;
    else
    if (!strncmp(lower, "child_label", 11))
        childType = XmCHILD_LABEL;
    else
    if (!strncmp(lower, "child_action", 12))
        childType = XmCHILD_ACTION;
    else
        badConversion = TRUE;

    XtFree(lower);

    if (badConversion)
        XtDisplayStringConversionWarning(_display, _from->addr, XmRSMBChildType);
    else
    {
        if (_to->addr == NULL)
            _to->addr = (XtPointer)&childType;
        else
        if (_to->size < sizeof(unsigned char))
            badConversion = TRUE;
        else
            *(unsigned char *)_to->addr = childType;
            _to->size = sizeof(unsigned char);
    }

    return !badConversion;
}

#ifdef _NO_PROTO
Widget XmCreateSmartMessageBox(_parent, _name, _warg, _numWarg)
Widget _parent;
char *_name;
ArgList _warg;
Cardinal _numWarg;
#else
Widget XmCreateSmartMessageBox(Widget _parent, char *_name, ArgList _warg, Cardinal _numWarg)
#endif /* _NO_PROTO */
{
	return (XtCreateWidget(_name, xmSmartMessageBoxWidgetClass, _parent, _warg, _numWarg));
}
